#include "OmSourceAddSourcesDlg.h"
#include "OmSourcePhotometryDlg.h"

#include <qapplication.h>
#include <qpushbutton.h>
#include <qdialog.h>
#include <qtable.h>
#include <qstringlist.h>
#include <qgroupbox.h>
#include <qlayout.h>
#include <qmessagebox.h>
#include <qtimer.h>
#include <qspinbox.h>
#include <qtooltip.h>
#include "OmSourceVersionNo.h"
#include "OmSourceParameters.h"
#include <iostream>
#include "OmSourceButton.h"
#include "OmSourceGroupBox.h"
#include "OmSourceLabel.h"
#include "OmSourceNumerics.h"
#include "OmImageDisplay.h"
#include "OmSourceRegions.h"
#include "OmSourceProcessDlg.h"
#include "average.h"
#include <cmath>

F90MODPROCPROTO(void,omsource_glue, deletebackgroundlist)();
F90MODPROCPROTO(void, omsource_glue, deletebackgroundregions)();
F90MODPROCPROTO(void, omsource_glue, initialisephotometry)();
F90MODPROCPROTO(void, omsource_glue, addregions1)(OmSourceRegions*);
F90MODPROCPROTO(void, omsource_glue, addbackgroundregions)(OmSourceRegions *,const float &);
F90MODPROCPROTO(void, omsource_glue, getthecounts)();
F90MODPROCPROTO(void, omsource_glue, appendtosourceregions)(OmSourceRegions *);
F90MODPROCPROTO(float, omsource_glue, getimagecounts)(float &, float&, float&);
F90MODPROCPROTO(void, omsource_glue, computesourceparameters)(OmSourceParameters &sources);
F90MODPROCPROTO(void, omsource_glue, computephotometry)(OmSourceParameters &sources);
F90MODPROCPROTO(void, omsource_glue, appendtosourcelist)(OmSourceParameters &sources);


extern F90MODVARDEF(OmSourceParameters, omsource_glue, mybackgroundsources);
extern F90MODVARDEF(OmSourceParameters, omsource_glue, mysources);
extern F90MODVARDEF(OmSourceParameters, omsource_glue, myaddedsources);

extern F90MODVARDEF(OmSourceRegions, omsource_glue, mybackgroundregions);

/********************************************************************************
 * Dialog for obtaining background regions from the user          
 *
 *******************************************************************************/
vector<float> AddSourcesDlg::_radiusList;
AddSourcesDlg::AddSourcesDlg(QWidget *parent,  OmImageDisplay *imageDisplay,
    float apertureRadius) :
    QDialog(parent, "", true), _imageDisplay(imageDisplay),
    _apertureRadius(apertureRadius)

{
 
    QString text("Please select new sources\n"
                 "by clicking the left mouse-button\n"
                 "The selected sources can be deleted\n"
                 "by clicking the 'Delete' button");
    _label = new OmSourceLabel(this, text, 20, 30, 220, 100);



    // Create the table for storing details about the selected background regions 
    _table = new QTable(10, 3, this, 0);
    _table->setGeometry(20, 150, 370, 120);
  
    int noColumns = 5;

    _table->setReadOnly(true);   // Can only read the table
    _table->setNumCols( noColumns);
    _table->setNumRows(1);   
   

    QHeader *th = _table->horizontalHeader();
    QToolTip::add(_table, "This table stores the following information about the new sources selected by you-\n\n"
                           "Column 1- the x coordinate (pixels)\n"
                           "Column 2- the y coordinate (pixels)\n"
                           "Column 3- the radius of the circle (pixels)\n"
                           "Column 4- the counts within a circular aperture having the radius in column 3\n"
		           "Column 5- the error in the counts"); 
   
     
    th->setLabel(0,"x coord");
    _table->setText(0, 0, "x coord");
 

    th->setLabel(1,"y coord");
    _table->setText(0, 1, "y coord");
 
    th->setLabel(2,"radius");
    _table->setText(0, 2, "radius");
 

    th->setLabel(3,"counts");
    _table->setText(0, 3, "counts");


    th->setLabel(4,"+/-");
    _table->setText(0, 4, "counts");


    /********************************************
    * Adjust the size of the cells
    *******************************************/
    for (int i = 0; i < noColumns; i++)
    {         
        _table->setColumnWidth(i, _table->text(0, i).length() * 10);
	_table->setText(0, i, " ");
    }
    _table->setNumRows(0);   
 
  
    OmSourceGroupBox *group = new OmSourceGroupBox( this, 
          "List operations", 20, 300, 120, 60, "An operation that is performed on the list of selected sources" );   
						  

    QHBoxLayout* layout = new QHBoxLayout(group);
    
    layout->setMargin(10);
    layout->setSpacing(10);  
   
 
    _deleteListBtn = new OmSourceButton(group, "Delete",
               "Click here to delete the source list");
  
    layout->addWidget( _deleteListBtn);
  
    connect(_deleteListBtn, SIGNAL(clicked()), this, SLOT (deleteList()) );


     /***********************************************************
     * Spin box to get the radius of the background aperture
     **********************************************************/
    int xPos = 20;
    int yPos = 370;
    int height = 30;

    OmSourceGroupBox *groupBox = new OmSourceGroupBox( this,  "Annulus (pixels)", xPos, yPos, 160, 70);
   
    xPos = 10;
    yPos = 20;
    _apertureTxt = new OmSourceLabel(groupBox, "Aperture",  xPos, yPos, height);
 
    xPos += 80; 
  
    _apertureSpinBox = new QSpinBox(groupBox, "SpinBox2" );
    _apertureSpinBox->setGeometry( xPos, yPos, 50, height ); 
    
    _apertureSpinBox->setMaxValue(30);
    _apertureSpinBox->setMinValue(3);
    _apertureSpinBox->setValue((int)_apertureRadius);

    connect(_apertureSpinBox, SIGNAL(valueChanged(int)), this, SLOT(apertureRadiusHasChanged(int)));
    
    text.sprintf("Change the background aperture radius\n (allowed values=%d-%d)", 
		 3, 30);
    QToolTip::add(_apertureSpinBox, text);

  
  

    QPushButton* okButton = new QPushButton("OK", this);
    QToolTip::add(okButton, "Click here when you have finished selecting sources\n"
		            "and wish to add them to the source-list.\n"
                            "Aperture photometry will be performed on each source-\n"
                            "If a large number of sources have been selected, or the\n"
		            "image is large, this may take some time!"); 
    okButton->setGeometry(100, 450, 100, 30);
    connect(okButton, SIGNAL( clicked() ), this, SLOT( accept1() ));
 

    QPushButton* cancelButton = new QPushButton("CANCEL", this);
    cancelButton->setGeometry(210, 450, 100, 30);
    connect(cancelButton, SIGNAL( clicked() ), this, SLOT( reject() ));
    QToolTip::add(cancelButton, "Click here when you have finished selecting sources\n"
		            "but do not wish to add them to the source-list");
  


    OmImageRegions::_noBackgroundRegions = 0;
  
    update();
 

    // Initialise some photometry arrays
    F90MODPROC(omsource_glue, initialisephotometry)();
 
   
    OmSourceRegions *points1 = OmImageRegions::GetNewRegions(_imageDisplay);
     
  
    if (points1) delete points1;	  
   

  
    _timer = new QTimer( this);

    connect( _timer, SIGNAL(timeout()), this, SLOT(updateRegions()) );
  
    _timer->start( 3000); // 2 seconds single-shot timer
     
         
};
/***************************************
 * Destructor
 **************************************/
AddSourcesDlg::~AddSourcesDlg()
{

    _timer->stop();
    delete _timer;

}
/****************************************************
 *
 ****************************************************/
void AddSourcesDlg::accept1()
{
  
    if (OmImageRegions::_noBackgroundRegions == 0) 
    {
  
      reject();
      return;
    }
 

  
    F90MODPROC(omsource_glue, appendtosourceregions)(OmImageRegions::_backgroundRegions);

    
    // Need to compute the parameters for each new source
     
  

    F90MODPROC(omsource_glue, addregions1)(OmImageRegions::_backgroundRegions);
 
 
        // Get the newly added sources 
    OmSourceParameters& sources = F90MODVAR(omsource_glue, myaddedsources);
    for (int i = 0; i < OmImageRegions::_noBackgroundRegions; i++)
        sources[i].photometryradius = _radiusList[i - 1];



    // Compute the photoemtry



    // Display the dialogue to do the photometry
    QString caption =  "Computing photometry- please wait";

    
    ProcessDlg *processDlg = new ProcessDlg(this, &sources, caption, 
            ProcessDlg::aperturePhotometry);
   
    processDlg->exec();

    delete processDlg;  


    //    F90MODPROC(omsource_glue, appendtosourcelist)(sources);


    //        F90MODPROC(omsource_glue, computephotometry)(sources);
  
 
    OmSourceParameters& mySources = F90MODVAR(omsource_glue, mysources);
    for (int i = 1, j = mySources.ubound() -  OmImageRegions::_noBackgroundRegions + 1; 
        i <= OmImageRegions::_noBackgroundRegions; i++, j++)
        mySources[j] = sources[i];
   
       
    F90MODPROC(omsource_glue, deletebackgroundlist)();

    delete OmImageRegions::_backgroundRegions;
   
    OmImageRegions::_noBackgroundRegions = 0;
    accept(); 
}
/**********************************************
 *
 *********************************************/
void AddSourcesDlg::deleteList()
{
   _timer->stop();

   _imageDisplay->send("regions deleteall");
	
   // Reset the table  
   _table->setNumRows(0);   
  
    _deleteListBtn->setDisabled(TRUE);
    // Delete the background regions

  
    F90MODPROC(omsource_glue, deletebackgroundlist)();
    F90MODPROC(omsource_glue, deletebackgroundregions)();
    delete OmImageRegions::_backgroundRegions;
  
 
    OmImageRegions::_noBackgroundRegions = 0;

    _radiusList.clear();
   
  

    
    OmSourceRegions *points1 = OmImageRegions::GetNewRegions(_imageDisplay);
    if (points1) delete points1;	  
 

    _timer->start( 5000);
 
     
}

void AddSourcesDlg::mousePressEvent( QMouseEvent *e )
{
    QApplication::beep();
    if ( e->button() == QMouseEvent::LeftButton )               // left button pressed
    QMessageBox::critical( 0, "omsource",
			       QString("LEFT MOUSE BUTTON")); 

}
/*********************************************************
* Checks the image for any new background regions
* and, if necessary, updates the image and table
********************************************************/
void AddSourcesDlg::updateRegions()
{ 
  

    // Get the new background regions
    OmSourceRegions *newRegions = OmImageRegions::GetNewRegions(_imageDisplay);
    
    if (newRegions == 0) return;
  

    if (newRegions->ubound() <= 0) return;
   
       
    OmImageRegions::updateBackgroundRegions(newRegions);
           
    // Compute the counts for each newly selected region     
    for (int i = newRegions->lbound(); i <= newRegions->ubound(); i++)
    {   
 
        float x = (*newRegions)[i].x();
        float y = (*newRegions)[i].y();

        float counts = F90MODPROC(omsource_glue, getimagecounts)(x, y, _apertureRadius);
	  
        float countsError = sqrt(counts);
  
        // Update the table
        updateTable(x, y, _apertureRadius, counts, countsError);  
        _radiusList.push_back(_apertureRadius);
    
       	
    }


    delete newRegions;  // Finished with the new background regions
 
   
    _imageDisplay->send("regions deleteall");
	
    _imageDisplay->displayCurrentSourceList();

    _imageDisplay->displaySourceRegions(OmImageRegions::_backgroundRegions, _radiusList);

        
}

/*****************************************************
 * Called when the value of apertureradius 
 *spin-box has changes
 ****************************************************/
void AddSourcesDlg::apertureRadiusHasChanged(int radius)
{
    _apertureRadius = radius;
    _apertureSpinBox->setValue(radius);

}
/***********************************************************
*
*
***********************************************************/
void AddSourcesDlg::updateTable(float x, float y, float radius, 
           float counts, float countsError)
{
    int rowNo = _table->numRows();
    
    _table->setNumRows(rowNo + 1);
    
    QString text;
                 
    text.sprintf("%6.2f", x);
    _table->setText(rowNo, 0, text);

    text.sprintf("%6.2f", y);
    _table->setText(rowNo, 1, text);

    text.sprintf("%2d",  (int)radius);
   _table->setText(rowNo, 2, text);

    text.sprintf("%8.2f", counts);  
   _table->setText(rowNo, 3, text);

    text.sprintf("%8.2f", countsError);  
   _table->setText(rowNo, 4, text);
 


}

/***********************************************************
* Loads all the new sources into the table
***********************************************************/
void AddSourcesDlg::update()
{

    if (OmImageRegions::_noBackgroundRegions == 0) return;
  
    _imageDisplay->displayCurrentSourceList();
    _imageDisplay->displaySourceRegions(OmImageRegions::_backgroundRegions, _radiusList);
  
 
    // Compute the counts for each newly selected region
    
    for (int i = OmImageRegions::_backgroundRegions->lbound(); 
           i <= OmImageRegions::_backgroundRegions->ubound(); i++)
    {   

        float x = (*OmImageRegions::_backgroundRegions)[i].x();
        float y = (*OmImageRegions::_backgroundRegions)[i].y();
       

           
	float counts = F90MODPROC(omsource_glue, getimagecounts)(x, y, _radiusList[i - 1]);
     
        float countsError = sqrt(counts);

          // Update the table
        updateTable(x, y, _radiusList[i - 1], counts, countsError);  
	

    }
 
} 








