! -*-f90-*-
! Originator:cbrindle
! Adapted:   Jorgo Bakker (jbakker@rssd.esa.int)
! Info:
!   Interactive program for selecting sources from an image
!   for further analysis
! Note:
!  * Should always call "omsource_init()" first, and "omsource_finish()" last.
!  * In comments, I usually talk about methods. A method may be either a
!    function OR subroutine within a module
!  * Any variable starting with 'my' is global to this module
   

module omsource_old
   use dal
   use cal
   use sasdal
   use oal
   use param
   use sla
   use errorHandling
   
   use frametime_mod, only : getFrameTime ! mssllib
   use getFilterID_mod
   use omutility_mod      ! In mssllib various fits access routines
   use omconstants_mod    ! In mssllib defines various constants
   use omsourceparam_mod  ! In mssllib defines Struct SourceParameters
   use omparameters_mod 
   use omphotometry_mod   ! In mssllib for doing aperture photometry
   use ommagnitude_mod    ! In mssllib for computing instrumental magnitude 
   use omsrctable_mod     ! In mssllib for creating the source-list table
   use omsourcecelestial_mod
   use omprogress_mod
   use omsourceregion_mod
   use omsourceparameter_mod      
   
   implicit none


   private ! all, but:

   public :: myImage
   public :: OmSourceRegion      ! type definition
   public :: ObsDetails           ! NB
   public :: ImageWindow          !

   public  :: omsource_init       ! subroutine ...()
   public  :: omsource_finish     ! subroutine ...()
   public  :: omsource_loadsources  ! subroutine ... ()
   public  :: omsource_delete     ! subroutine ...()
   public  :: omsource_deleteList 

   public  :: omsource_replace  ! subroutine ...()

   public  :: omsource_recompute  ! subroutine ...()
   public  :: omsource_add        ! subroutine ...()
   public  :: omsource_add1        ! subroutine ...()
   public  :: omsource_computeparameters
   public  :: omsource_photometry
   public :: omsource_getbackgroundmode

   public  :: omsource_addBackgroundSources

   public  :: omsource_getnewsources      ! subroutine ...()

 
   interface construct
      module procedure construct_od
   end interface

   interface destruct
      module procedure destruct_od
   end interface
 
   interface construct
      module procedure construct_iw
   end interface

   interface destruct
      module procedure destruct_iw
   end interface


   interface construct
      module procedure construct_celestial
   end interface

   interface destruct
      module procedure destruct_celestial
   end interface


contains
!*********************************************************************
! Info: Initialises the fortran part of omsource.
!       It loads the image pdata and properties, and -if any- it also
!       creates a source list from an already existing sources dataset
!*********************************************************************
subroutine omsource_init(sources, backgroundSources, &
                         addedSources, addedBackgroundSources, &
                         photometricParameters)

      type(SourceParameters),  pointer    :: sources(:)
      type(SourceParameters),  pointer    :: backgroundSources(:)
      type(SourceParameters),  pointer    :: addedSources(:)
      type(SourceParameters),  pointer    :: addedBackgroundSources(:)


      type(PhotometryParameters), intent(out) :: photometricParameters
      
!      type(OmProgressT)        :: reporter    ! ... of progress to GUI
      character(*) , parameter :: routine="omsource_old.omsource_init"
      real(single) :: sigma 
      nullify(sources)       ! proper state
      nullify(backgroundSources)
      nullify(addedSources)
      nullify(addedBackgroundSources)

      ! --- initialise globals in this module


      myDeadFraction       = -1.
      myFrameTime          = -1.
      myDetectionRate      = -1.
      myDetectionMagnitude = -1.

      nullify(myImage)
      nullify(myQuality)

   
      ! --- extract the image from the file
      myImageSet = dataSet(stringParameter("imageset"), modify)

      ! load the image parameters (WINDOWX, etc) from the image fits file...
      call construct(myImageWindow, set = myImageSet)


      ! set the frame time etc...


      myFrameTime = myImageWindow%frameTime
      myDeadFraction = myImageWindow%deadFraction
      frameTime = myFrameTime
      deadFrac = myDeadFraction

!      call setProgress(reporter,4)



    ! Get the observation details for plotting
      call construct(myObsDetails, set = myImageSet)

      !<TODO: missing DATATYPE in dataset, otherwise we could do this:
      !call CAL_setState(myImageSet)
      call CAL_setState(OM, filterId = filterIdToInt(trim(myObsDetails%filter)))
      !call setProgress(reporter,5)
      !TODO>

      ! create celestial information; also sets state of CAL
      call construct(myCelestial, set=myImageSet)
!      call setProgress(reporter, 6)

      ! now init radii once, and once only, and bin them now
      myApertureRadius = 12.0                ! NEED THIS IN UNBINNED PIXELS

      myCalRadius = CAL_omGetApertureRadius() ! NEED THIS IN UNBINNED PIXELS


      if (myCalRadius .gt. myApertureRadius) then
        !**********************************************
        ! Remember that counts have to be extrapolated.
        !**********************************************
        myScaleCounts = .true.
        myScaleCountsRatio = (myCalRadius / myApertureRadius) ** 2 
      end if

      photometricParameters%apertureRadius =  myApertureRadius / real(myImageWindow%bx)
    
      photometricParameters%calApertureRadius =  myCalRadius /  real(myImageWindow%bx)
      photometricParameters%binX = myImageWindow%bx
     

      myBgrRadiusIn    = photometricParameters%calApertureRadius + 2.0

      myBgrRadiusOut   = sqrt(3.0 *   photometricParameters%calApertureRadius * &
          photometricParameters%calApertureRadius +  myBgrRadiusIn  * myBgrRadiusIn)
  
      !************************************* 
      ! Get the Plate scale (arcsecs/pixel)
      !*************************************
      call CAL_omGetPlateScale(0.55_double, myImageWindow%plateScale)
     
!      call setProgress(reporter, 7)

      ! Get the RA and DEC of the image centre
!      call setLabel(reporter,"Computing sky coordinates")
      call getCelestialCentre(myCelestial, myImageWindow%RA, myImageWindow%Dec)
!      call setProgress(reporter, 8)

      ! Get the pixels
      call createImageArray(myImageSet, myImage)

    ! Compute the background mode
      myBgrMode = backgroundMode(myImage, size(myImage, 1), &
      size(myImage, 2), 2.0, sigma) 





!      call setProgress(reporter, 9)

      ! Get the quality of the pixels
      call createQualityArray(myImageSet, myQuality)
!      call setProgress(reporter, 10)
!      call destruct(reporter)
      ! --- finished general information


      ! Load the old sourcelist file if one has been given on the
      ! command line
  
  
      call loadSources(sources, stringParameter("oldsrclistset"))

      mySourcesAreLoaded = associated(sources)
     
   end subroutine omsource_init

!*******************************************************************************
! Info: Loads the sources from the source-list file produced by omdetect & ommag
!       The 'sources' argument is populated with the detected sources.
!*******************************************************************************
subroutine omsource_loadsources(sources, sourceListFile)
    type(SourceParameters),  pointer    :: sources(:)
    
    character(256), intent(in) :: sourceListFile;
 
 
    call loadSources(sources, sourceListFile)
     
     
    ! Compute the RA and Declination of each source

!      call setLabel(reporter,"Celestial properties")      
    call setCelestial(myCelestial, sources)

end subroutine omsource_loadsources

!*************************************************************
! omsource_recompute
!
!
! Info: recomputes the properties for a source=sources(id)
!***************************************************************
subroutine omsource_recompute(sources, id)
    type(SourceParameters),  pointer    :: sources(:)
    integer,                  intent(in) :: id

    integer(int32) :: nx, ny
    type(SourceParameters), pointer :: s
    type(SourceParameters), pointer  :: tmp(:) => null()
    integer(int32)                  :: bin(2)
    character(*) , parameter :: routine="omsource_old.omsource_recompute"
    integer(int16), pointer :: levelImage(:, :) => null()
    real(single) :: corrBackgroundCountRateErr = 0.0


    s=>sources(id)

    nx = size(myImage, 1)
    ny = size(myImage, 2)

    bin= (/myImageWindow%bx, myImageWindow%by/)
 

!   if (s%apertureRadius .le. 0.0) s%apertureRadius = myApertureRadius

!write(*, *) "omsource_old line 279 s%apertureRadius=", s%apertureRadius
!write(*, *) "My aperture Radius=", myApertureRadius 

    ! Do aperture photometry on  source
    allocate(tmp( 1 : 1))
    tmp(1) = s
!    tmp(1)%extensionFlag = 0 ! TODO: why here, and why not before?


  
!    s = tmp(1)



    levelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "LevelImage")

!call fatal("stopped")
!    call AperturePhotometry(nx, ny, myImage, levelImage, &
!        myQuality, myApertureRadius,         &  
!        size(sources), sources)                  

!write(*, *) "Line 300 calling aperturephotometry - aperture radius=",  myapertureRadius / real(binX)
!write(*, *) "backgroundcountrate=", tmp(1)%rawbackgroundcountrate
    call AperturePhotometry(nx, ny, myImage, levelImage, &
        myQuality, myapertureRadius / real(binX),         &  
        1, tmp)                  



   s = tmp(1)


    s%apertureRadius = s%apertureRadius / real(binX)
    

    deallocate(levelImage)
    nullify(levelImage)
!write(*, *) "omsource_old.f90 line 300 calling ComputeMagnitude"
!write(*, *) "corrected count rate=", s%corrCountRate
!write(*, *) "corrected background count rate=", s%corrBackgroundCountRate
!write(*, *) "SourceAperture radius=",  s%apertureRadius

!write(*, *) "Aperture radius=",  myApertureRadius
!write(*, *) "cal radius=",  myCalRadius
!write(*, *) "psf1 correction=", s%psf1Correction
!write(*, *) "coincidence-loss correction=", s%coincidenceLossCorr

    call ComputeMagnitude(s%corrCountRate, s%corrCountRateErr, &
                       s%corrBackgroundCountRate, corrBackgroundCountRateErr, &
                       s%xPos, s%yPos,             &
                       myScaleCounts, s%extensionFlag, &
                       myCalRadius, myApertureRadius, &
                       myFrameTime, myDeadFraction, &
                       s%magnitude, s%magnitudeErr, s%psf2Correction)
!write(*, *) "After calling ComputeMagnitude"
!write(*, *) "corrected count rate=", s%corrCountRate
!write(*, *) "corrected background count rate=", s%corrBackgroundCountRate
!write(*, *) "Magnitude=", s%magnitude
!write(*, *) "psf2 correction=", s%psf2Correction
    sources(id) = s
    deallocate(tmp)
    nullify(tmp)
end subroutine omsource_recompute
!*************************************************************
! omsource_delete
!
!
! Info: deletes source=sources(id) from the source file; then all sources are
!       renumbered to make the IDs consitent.
!************************************************************/
subroutine omsource_delete(sources, id)
    type(SourceParameters),  pointer    :: sources(:)
    integer,                  intent(in) :: id

    type(SourceParameters),  pointer :: v(:)
    integer                           :: l, h  ! lbound/ubound
    integer                           :: idx  ! iterator
    integer                           :: off  ! 

      ! assume that sources(id) exists
    l = maxval(lbound(sources))
    h = maxval(ubound(sources))
    off = 0

    call new(v, low = l, high = h - 1)
    do idx = l, h
        if (idx .eq. id) then
            off = -1
            cycle
        end if
        v(idx + off) = sources(idx)
    end do

    call delete(sources)
    sources => v
    nullify(v)

end subroutine omsource_delete
!***********************************************
! Deletes the whole source list
!**********************************************
subroutine omsource_deleteList(sources)

    type(SourceParameters),  pointer    :: sources(:)

    if (.not. associated(sources)) return
    call delete(sources)   
    nullify(sources)


    call new(sources, low = 1, high = 0)

end subroutine omsource_deleteList

!**************************************************************
! Info: add regions to the source list
! Note: this is never called when no regions were available!
!
!**************************************************************
subroutine omsource_add(sources, regions, radius, newSources)
    type(SourceParameters),  pointer    :: sources(:)
    type(OmSourceRegion),     intent(in) :: regions(:)
    real(single), intent(in) :: radius
    type(SourceParameters),  pointer    :: newSources(:)

    ! locals
    integer                         :: i
    integer,                pointer :: idx(:)
    type(SourceParameters),pointer :: tmp(:)
    type(SourceParameters),pointer :: added(:)
    type(OmProgressT)               :: reporter    ! ... of progress to GUI

    call construct(reporter, label = "Compute added sources...", steps = 4)
    call setProgress(reporter,1)

    
    call omsource_deleteList(newSources)
       
   
    ! --- check new regions against current sources
    nullify(added)
    if (associated(sources)) then
        idx => getNewRegions(regions, sources, real(16 / myImageWindow%bx))
        if (associated(idx)) then
            call new(added, low = 1, high = size(idx))

            do i = 1, size(idx)
                added(i)%xPos=   regions(idx(i))%x
                added(i)%yPos=   regions(idx(i))%y
                added(i)%exposure= myObsDetails%exposure
                added(i)%apertureRadius = radius
            end do

         
              deallocate(idx)
    
       end if
       nullify(idx)
    else
        call new(added,low = 1, high = size(regions))

        do i = 1, size(regions)
            added(i)%xPos=   regions(i)%x
            added(i)%yPos=   regions(i)%y
            added(i)%exposure= myObsDetails%exposure
     
  

        end do
      
    end if
    call setProgress(reporter,2)

    ! compute properties for new sources
    if (associated(added)) then


        call ComputeParameters(added)
        call new(newSources, low = 1, high = size(added))
        newSources = added
      

    
    end if
    call setProgress(reporter, 3)



    ! merge
    if (associated(sources) .and. associated(added)) then
        
        tmp=>sources

        call new(sources, low = 1, high = size(tmp) + size(added))
        sources(1 : size(tmp)) = tmp
        do i = 1, size(added)
            sources(size(tmp) + i) = added(i)
        end do
        call delete(tmp)
        call delete(added)
    else
        ! either added or sources did not exists, so:
        if (associated(added)) sources=>added
        nullify(added)
    end if

    call setProgress(reporter,4)
    call destruct(reporter)

end subroutine omsource_add


!**************************************************************
! Info: add regions to the source list
! Note: this is never called when no regions were available!
!
!**************************************************************
subroutine omsource_add1(sources, regions, newSources)
    type(SourceParameters),  pointer    :: sources(:)
    type(OmSourceRegion),     intent(in) :: regions(:)
    type(SourceParameters),  pointer    :: newSources(:)

    ! locals
    integer                         :: i
  
    type(SourceParameters),pointer :: tmp(:)
    type(SourceParameters),pointer :: added(:)
    type(OmProgressT)               :: reporter    ! ... of progress to GUI

    call construct(reporter, label = "Compute added sources...", steps = 4)
    call setProgress(reporter,1)

    
    call omsource_deleteList(newSources)
       
  
    ! --- check new regions against current sources
    nullify(added)
 
    call new(added, low = 1, high = size(regions))

    do i = 1, size(regions)
        added(i)%xPos=   regions(i)%x
        added(i)%yPos=   regions(i)%y
        added(i)%exposure= myObsDetails%exposure
       
      
    end do
      
   
    call setProgress(reporter,2)

    ! compute properties for new sources
    if (associated(added)) then

        ! Compute the source positions, etc
     

        call ComputeSourcePositions(added)
   
    
        ! Compute RAs and decs
        call setCelestial(myCelestial, added)
        call new(newSources, low = 1, high = size(added))
        newSources = added
        

    
    end if
    call setProgress(reporter, 3)



    ! merge
    if (associated(sources) .and. associated(added)) then
        
        tmp=>sources

        call new(sources, low = 1, high = size(tmp) + size(added))
        sources(1 : size(tmp)) = tmp
        do i = 1, size(added)
            sources(size(tmp) + i) = added(i)
        end do
        call delete(tmp)
        call delete(added)
    else
        ! either added or sources did not exists, so:
        if (associated(added)) sources=>added
        nullify(added)
    end if
   

    call setProgress(reporter,4)
    call destruct(reporter)

end subroutine omsource_add1






!**************************************************************
! Info: add regions to the background source list
! ignoring "sources"
! Note: this is never called when no regions were available!
!
!**************************************************************
subroutine omsource_addBackgroundSources(sources, &
        backgroundSources, regions, apertureRadius)
    type(SourceParameters),  pointer    :: sources(:)
    type(SourceParameters),  pointer    :: backgroundSources(:)

    type(OmSourceRegion),     intent(inout) :: regions(:)
    real(single), intent(in) :: apertureRadius
    ! locals
    integer                         :: i, j, k, k1
    integer,                pointer :: idx(:)
    real(single)  :: radius
    type(SourceParameters),pointer :: tmp(:)
    type(SourceParameters),pointer :: added(:)
    type(OmProgressT)               :: reporter    ! ... of progress to GUI

    call construct(reporter, label = "Compute added sources...", steps = size(regions))
    call setProgress(reporter,1)

    ! --- check new regions against current sources
    nullify(added)
    if (associated(backgroundSources)) then
        idx => getNewBackgroundRegions(regions, sources, backgroundSources, real(16 / myImageWindow%bx))


        if (associated(idx)) then
            call new(added, low = 1, high = size(idx))
            do i = 1, size(idx)
                call setProgress(reporter, i)

                added(i)%xPos = regions(idx(i))%x
                added(i)%yPos = regions(idx(i))%y
                added(i)%exposure = myObsDetails%exposure   
                
                added(i)%apertureRadius = apertureRadius;
                added(i)%semia = apertureRadius;
                added(i)%semib = apertureRadius;
                added(i)%theta = 0.0;
        

            end do
            deallocate(idx)
        end if
        nullify(idx)
    else
     
 
        iLoop : do i = 1, size(regions)
            k = 0            
            do j = 1, size(regions)
                if (i == j) cycle
                radius = sqrt((regions(i)%x - regions(j)%x) ** 2 + (regions(i)%y - regions(j)%y) ** 2)
                if (radius < 1.0) cycle iLoop
                  
             end do 
             k = k + 1
             regions(k) = regions(i)
        end do iLoop



       

        if (associated(sources)) then
            k1 = 0
            iLoop1 : do i = 1, k 
                do j = 1, size(sources)
                    radius = sqrt((regions(i)%x - sources(j)%xpos) ** 2 + (regions(i)%y - sources(j)%xpos) ** 2)
                    if (radius < 1.0) cycle iLoop1
                end do
                k1 = k1 + 1
                regions(k1) = regions(i)    
            end do iLoop1  
            k = k1
        end if
        
        if (k > 0) then
            call new(added, low = 1, high = k)
            do i = 1, k
            
                added(i)%xPos = regions(i)%x
                added(i)%yPos = regions(i)%y
                added(i)%exposure = myObsDetails%exposure
 
                added(i)%apertureRadius = apertureRadius;
                added(i)%semia = apertureRadius;
                added(i)%semib = apertureRadius;
                added(i)%theta = 0.0;
     


             end do
         end if
    end if
  

    ! compute properties for new sources
    if (associated(added)) then


!        call ComputeParameters(myImage, added)
        
    end if
    call setProgress(reporter, 3)

    ! merge
    if (associated(backgroundSources) .and. associated(added)) then
        
        tmp=>backgroundSources

        call new(backgroundSources, low = 1, high = size(tmp) + size(added))
        backgroundSources(1 : size(tmp)) = tmp
        do i = 1, size(added)
            backgroundSources(size(tmp) + i) = added(i)
        end do
        call delete(tmp)
        call delete(added)
    else
        ! either added or sources did not exists, so:
        if (associated(added)) backgroundSources=>added
        nullify(added)
    end if

    call setProgress(reporter,4)
    call destruct(reporter)

end subroutine omsource_addBackgroundSources

!**************************************************************
! Info: gets the new regions
! 
!
!**************************************************************
subroutine omsource_getnewsources(sources, regions, newSources)
    type(SourceParameters),  pointer    :: sources(:)
    type(OmSourceRegion),     intent(in) :: regions(:)
    type(SourceParameters),  pointer    :: newSources(:)


    ! locals
    integer                         :: i
    integer,                pointer :: idx(:)
   
  
   

    ! --- check new regions against current sources
    if (associated(newSources)) call delete(newSources)

    nullify(newSources)

    if (associated(sources)) then
    

        idx => getNewRegions(regions, sources, real(16 / myImageWindow%bx))
        if (associated(idx)) then
            call new(newSources, low = 1, high = size(idx))
            do i = 1, size(idx)
                newSources(i)%xPos=   regions(idx(i))%x
                newSources(i)%yPos=   regions(idx(i))%y

                
                newSources(i)%exposure= myObsDetails%exposure
                newSources(i)%apertureRadius = 0.0 ! myApertureRadius;
                newSources(i)%semia = myApertureRadius;
                newSources(i)%semib = myApertureRadius;
                newSources(i)%theta = 0.0;
              
            end do

  


            deallocate(idx)
        end if
        nullify(idx)
    else
        call new(newSources, low = 1, high = size(regions))
        do i = 1, size(regions)
            newSources(i)%xPos =   regions(i)%x
            newSources(i)%yPos =   regions(i)%y
            newSources(i)%exposure = myObsDetails%exposure
            newSources(i)%semia = myApertureRadius;
            newSources(i)%semib = myApertureRadius;
            newSources(i)%theta = 0.0;
           
        end do
    end if



end subroutine omsource_getnewsources

!******************************************************
!
!*****************************************************
subroutine omsource_replace(id, source, sources)
    integer, intent(in) :: id
    type(SourceParameters), intent(in) :: source
    type(SourceParameters),  pointer    :: sources(:)
      
    sources(id) = source

end subroutine omsource_replace 
!******************************************************
! Info: saves sources, and frees memory
! used globals:
!   myCelestial
!   myImage
!   myImageSet
!   myImageWindow
!   myObsDetails
!   myQuality
!******************************************************
subroutine omsource_finish(sources, oldSources, backgroundSources, &
                           addedSources, addedBackgroundSources    )
    type(SourceParameters),  pointer    :: sources(:)
    type(SourceParameters),  pointer    :: oldSources(:)

    type(SourceParameters),  pointer    :: backgroundSources(:)
    type(SourceParameters),  pointer    :: addedSources(:)
    type(SourceParameters),  pointer    :: addedBackgroundSources(:)


    ! Add the sources to a table in the new source-list file
    call omsource_save(sources,&
                       stringParameter("newsrclistset"),&
                       myImageSet,&
                       myImageWindow,&
                       myObsDetails)
    ! Close the image set, as we don't need it anymore.
    call release(myImageSet)

    ! Free memory for this module globals
    call destruct(myCelestial)
    call destruct(myImageWindow)
    call destruct(myObsDetails)
    if (associated(myImage))   deallocate(myImage)
    if (associated(myQuality)) deallocate(myQuality)

    ! And the sources themselves
    call delete(sources)
    call delete(oldSources)
    
    call delete(backgroundSources)
    call delete(addedSources)
    call delete(addedBackgroundSources)
  


end subroutine omsource_finish



!*******************************************************************************
!--------------- helper functions private to this module -------------------
!
!
! Info: Compares the regions against the existing sources, and returns a 
!       vector containing indices to unique regions
! Note: * looks like the old MatchPoints subroutine, but cleaned-up
!       * returns a pointer, that must be destroyed by the caller
!********************************************************************************
function getNewRegions(r, s, toleranceIn) result(idx)
    type(OmSourceRegion),   intent(in)    :: r(:)
    type(SourceParameters), intent(in)    :: s(:)
    integer,                pointer       :: idx(:)
    real(single), optional, intent(in)    :: toleranceIn

    ! locals
    real(single)                          :: radius, tolerance
    integer                               :: i, j, nToKeep
    logical                               :: keep(size(r))
 

    ! init
    nullify(idx) ! always be on the safe side ;-)
    if (present(toleranceIn)) then
        tolerance = toleranceIn
    else
        tolerance = 5.0
    end if
    nToKeep = size(keep)

  

   ! populate the 'keep' and count the unique points
    do i = 1, size(r)
        keep(i) = .true.

        do j = 1, size(s)
            radius = sqrt((s(j)%xPos - r(i)%x) ** 2 +  (s(j)%yPos - r(i)%y)** 2)
            if (radius .le. tolerance) then ! overlap
               keep(i) = .false.
               nToKeep = nToKeep - 1

               exit ! inner loop
            end if
        end do
    end do
    if (nToKeep .eq. 0) return

    ! now we create an index of all regions that are new
    allocate(idx(nToKeep))
!    call assertOn(status/=0,"omsource_old.getNewRegions.idx")

    j = 0
    do i = 1, size(r)
        if (.not. keep(i)) cycle
        j = j + 1
        idx(j) = i
    end do
end function getNewRegions






!*******************************************************************************
!--------------- helper functions private to this module -------------------
!
!
! Info: Compares the regions against the existing sources, and returns a 
!       vector containing indices to unique regions
! Note: * looks like the old MatchPoints subroutine, but cleaned-up
!       * returns a pointer, that must be destroyed by the caller
!********************************************************************************
function getNewBackgroundRegions(region, source, background, toleranceIn) result(idx)
    type(OmSourceRegion), intent(in)      :: region(:)
    type(SourceParameters), pointer    :: source(:)
    type(SourceParameters), pointer    :: background(:)

    integer,                pointer       :: idx(:)
    real(single), optional, intent(in)    :: toleranceIn

    ! locals
    real(single)                          :: radius, tolerance
    integer                               :: i, j, nToKeep
    logical                               :: keep(size(region))
    logical, pointer                      :: identicalSource(:) => null()
    integer                               :: status


    ! init
    nullify(idx) ! always be on the safe side ;-)
    if (present(toleranceIn)) then
        tolerance = toleranceIn
    else
        tolerance = 5.0
    end if
    nToKeep = size(keep)



    ! populate the 'keep' and count the unique points
    if (associated(source)) then

    
        allocate(identicalSource(1 : size(source)))
        identicalSource = .false.
        ! Check for 2 identical sources




        do i = 1, size(source)
            if (identicalSource(i)) cycle
            
            do j = 1, size(source)
                
                if (i .ne. j .and. .not. identicalSource(j)) then
                    radius = sqrt((source(j)%xPos - source(i)%xPos) ** 2 +  (source(j)%yPos - source(i)%yPos)** 2)

                    if (radius .le. tolerance) then ! overlap
                        identicalSource(j) = .true.  ! Just keep source i
                       
                    end if
                end if
            end do
        end do
    
  

        do i = 1, size(region)
                
            keep(i) = .true.

            do j = 1, size(source)
                if (identicalSource(j)) then
 
                    keep(i) = .false.
                    nToKeep = nToKeep - 1

                    exit ! inner loop
                end if 
                radius = sqrt((source(j)%xPos - region(i)%x) ** 2 +  (source(j)%yPos - region(i)%y)** 2)
                if (radius .le. tolerance) then ! overlap
                    keep(i) = .false.
                    nToKeep = nToKeep - 1

                    exit ! inner loop
                end if
            end do
        end do
        deallocate(identicalSource)
    end if

    

    allocate(identicalSource(1 : size(background)))
    identicalSource = .false.
    ! Check for 2 identical sources
    do i = 1, size(background)
        if (identicalSource(i)) cycle
            
        do j = 1, size(background)
                
            if (i .ne. j .and. .not. identicalSource(j)) then
                radius = sqrt((background(j)%xPos - background(i)%xPos) ** 2 + &
                        (background(j)%yPos - background(i)%yPos)** 2)

                if (radius .le. tolerance) then ! overlap
                    identicalSource(j) = .true.  ! Just keep source i
                       
                end if
            end if
        end do
    end do


    do i = 1, size(region)

        if (.not. keep(i)) cycle

        do j = 1, size(background)


            if (identicalSource(j)) then
 
                keep(i) = .false.
                nToKeep = nToKeep - 1

                exit ! inner loop
            end if 
         

            radius = sqrt((background(j)%xPos - region(i)%x) ** 2 +  (background(j)%yPos - region(i)%y)** 2)
            if (radius .le. tolerance) then ! overlap
               keep(i) = .false.
               nToKeep = nToKeep - 1

               exit ! inner loop
            end if
        end do
    end do

    deallocate(identicalSource)

    if (nToKeep .eq. 0) return

    ! now we create an index of all regions that are new
    allocate(idx(nToKeep), stat = status)
!    call assertOn(status/=0,"omsource_old.getNewRegions.idx")

    j = 0
    do i = 1, size(region)
        if (.not. keep(i)) cycle
        j = j + 1
        idx(j) = i
    end do


end function getNewBackgroundRegions

!**************************************************
! Gets the 2-d image from the fits file
!**************************************************
subroutine createImageArray(dataSet,image)
    type(dataSetT), intent(in) :: dataSet
    real(single),   pointer    :: image(:, :)

    type(ArrayT)           :: handle
    integer(int8), pointer :: int8ImageP(:, :)     ! Integer8 array pointer.
    integer(int32),pointer :: int32ImageP(:, :)    ! Integer32 array pointer.
    integer(int16),pointer :: int16ImageP(:, :)    ! Integer16 array pointer.
    real(single),  pointer :: real32ImageP(:, :)   ! Real32 array pointer.
    real(double),  pointer :: real64ImageP(:, :)   ! Real64 array pointer.
    integer                :: sz(2)
    integer                :: status

    nullify(image)
    handle = array(dataSet, "PRIMARY", read)

    sz = dimensions(handle)
    if (all(sz > 0)) then
        allocate( image(0 : sz(1) - 1, 0 : sz(2) - 1), stat = status)
!        call assertOn(status/=0,"omsource_old.createImageArray.image")

        select case(arrayDataType(handle))
        case(integer8)
            int8ImageP => int8Array2Data(handle)
            image = real(int8ImageP)

        case(integer16)
            int16ImageP => int16Array2Data(handle)
            image = real(int16ImageP)

        case(integer32)
            int32ImageP => int32Array2Data(handle)
            image = real(int32ImageP)

        case(real32)
            real32ImageP => real32Array2Data(handle)
            image = real32ImageP

        case(real64)
            real64ImageP => real64Array2Data(handle)
            image = real64ImageP
        end select
    end if

    call release(handle)
end subroutine createImageArray
!*********************************************************
!
!*********************************************************
subroutine createQualityArray(dataSet,quality)
    type(dataSetT), intent(in) :: dataSet
    logical(bool),  pointer    :: quality(:, :)

    type(ArrayT)                :: handle
    integer(int8), pointer :: arrayP(:, :)
    integer(int32)              :: sz(2)
    integer                     :: status

    nullify(quality)
    handle = array(dataSet, "QUALITY", read)

    sz = dimensions(handle)
    if (all(sz > 0)) then
        arrayP => int8Array2Data(handle)

        allocate( quality(0 : sz(1) - 1, 0 : sz(2) - 1), stat=status)
!        call assertOn(status/=0,"omsource_old.createQualityArray.quality")

        where (arrayP .ne. 0)
            quality = .true.
        elsewhere
            quality = .false.
        end where
    end if

    call release(handle)
end subroutine createQualityArray


!*********************************************************************************
! subroutine loadSources
! Info: loads sources found in dataset(setName), and returns a newly created vector
!       of sources.
! 
!********************************************************************************
subroutine loadSources(sources, setName)

    type(SourceParameters), pointer    :: sources(:)
    character(len=*), intent(in) :: setName
   
  
    ! locals
    integer(int32)        :: nSources
  
    type(DataSetT)        ::  set
    type(TableT)          :: inTableH
    real(single)     :: innerBackgroundRadius
    real(single)     :: outerBackgroundRadius
         
   
    type(OmProgressT)     :: reporter    ! ... of progress to GUI    


      ! --- don't bother if none available
    if (trim(setName) == "" ) return

    ! Read in the OM OSW Source List file.
    set = dataSet(setName, read)
    myExposure = GetReal32Attribute(set, "EXPOSURE", 0.0)  

    if (.not. hasBlock(set, "SRCLIST")) then
        call destruct(reporter)

        ! Incompatible file - can't continue
        write (errstr, *) "Could not find any compatibile blocks in set ", trim(setName)
        call fatal("BadFileType")

    else
        ! Get a handle on the source list binary table extension.
        inTableH = table(set, "SRCLIST")

    end if

     

    ! Get the limiting magnitude of the image
    myDetectionRate      = GetReal32Attribute(set, "CRLIMIT")
    myDetectionMagnitude = GetReal32Attribute(set, "MAGLIMIT")
    
    
    innerBackgroundRadius =  GetReal32Attribute(set, "INNBCKAN")
    outerBackgroundRadius =  GetReal32Attribute(set, "OUTBCKAN")

    if (innerBackgroundRadius .eq. 0.0) then

        innerBackgroundRadius = myBgrRadiusIn
        outerBackgroundRadius =  myBgrRadiusOut

    end if
 
    call construct(reporter, label="Loading sourcelist data", steps = 23)



    ! Get the number of rows in the source list binary table extension.
    nSources = numberOfRows(inTableH)

    if (nSources == 0) then
        call destruct(reporter)

        call release(inTableH)
        call release(set)
        return

    end if

    call delete(sources)
    call new(sources, low = 1, high = nSources)
 
    ! Get the data for each source from the source list file
    call setProgress(reporter, 1)
  
    call LoadReal32Column(inTableH, "XPOS",  sources%xPos)
    call setProgress(reporter, 2)
  
    call LoadReal32Column(inTableH, "YPOS",  sources%yPos)
    call setProgress(reporter, 3)
  
    call LoadReal32Column(inTableH, "FWHM_MAJ",  sources%semiA)
    call setProgress(reporter, 4)
  
    call LoadReal32Column(inTableH, "FWHM_MIN",  sources%semiB)
    call setProgress(reporter, 5)
  
    call LoadReal32Column(inTableH, "PA", sources%theta)
    call setProgress(reporter, 6)
  
    call LoadReal32Column(inTableH, "RATE", sources%rawCountRate)
    call setProgress(reporter, 7)
  
    call LoadReal32Column(inTableH, "RATE_ERR", sources%rawCountRateErr)
    call setProgress(reporter,8)

    call LoadReal32Column(inTableH, "CORR_RATE", sources%corrCountRate)
    call setProgress(reporter, 9)
  
    call LoadReal32Column(inTableH, "CORR_RATE_ERR", sources%corrCountRateErr)
    call setProgress(reporter, 10)
  
    call LoadReal32Column(inTableH, "COINCLOSS_CORR", sources%coincidenceLossCorr)
    call setProgress(reporter,11)
    call LoadReal32Column(inTableH, "PSF1_CORR", sources%psf1Correction)
    call setProgress(reporter, 12)
  
    call LoadReal32Column(inTableH, "PSF2_CORR", sources%psf2Correction)
    call setProgress(reporter,13)

    call LoadReal32Column(inTableH, "BACKGROUND_RATE",  sources%rawBackgroundCountRate)
    call setProgress(reporter, 14)
  
    call LoadReal32Column(inTableH, "CORR_BACKGROUND_RATE",  sources%corrbackgroundCountRate)
    call setProgress(reporter,15)

    call LoadReal32Column(inTableH, "MAG", sources%magnitude)
    call setProgress(reporter, 16)
  
    call LoadReal32Column(inTableH, "MAGERR", sources%magnitudeErr)
    call setProgress(reporter, 17)
  
    call LoadReal32Column(inTableH, "SIGNIFICANCE", sources%significance)
    call setProgress(reporter, 18)
  
    call LoadReal32Column(inTableH, "APERTURE", sources%apertureRadius)


    call LoadInt8Column(inTableH, "EFLAG", sources%extensionFlag)
    
    ! Need binned aperture radius
    where(sources%extensionflag .eq. 1) 
        sources%apertureRadius = 0.0
    elsewhere
        sources%apertureRadius = sources%apertureRadius / myImageWindow%bx
    endwhere
    call setProgress(reporter, 19)

    ! Get the RAS and decs if they exist
    call LoadReal64Column(inTableH, "RA", sources%rightAscension)
    call setProgress(reporter, 20)
  
    call LoadReal64Column(inTableH, "DEC", sources%declination)
    call setProgress(reporter, 21)
  
    call LoadInt8Column(inTableH, "EFLAG", sources%extensionFlag)


    call setProgress(reporter,22)

    call release(inTableH)
    call release(set)


    sources%innerBckRadius = innerBackgroundRadius

    sources%outerBckRadius = outerBackgroundRadius
    sources%exposure = myExposure 


      ! Convert width to arcsecs
  
    sources%semiA = sources%semiA / ( myImageWindow%bx *  myImageWindow%plateScale * 2.354)
    sources%semiB = sources%semiB / ( myImageWindow%bx *  myImageWindow%plateScale * 2.354)
         
    call setProgress(reporter,23)

    call destruct(reporter)

end subroutine loadSources

!**********************************************************
! Info: saves the sources into dataset(fileName)
! Used globals:
!   mySourcesAreLoaded
!   myDetectionRate
!   myDetectionMagnitude
!   myFrameTime
!   myDeadFraction
!   myNullValue
! TODO: save unbinnend radius? save baground radii?
!*********************************************************
subroutine omsource_save(sources, fileName, imageSet, iw, od)
    type(SourceParameters), pointer    :: sources(:)
    character(*),            intent(in) :: fileName
    type(DataSetT),          intent(in) :: imageSet
    type(ImageWindow),       intent(in) :: iw
    type(ObsDEtails),        intent(in) :: od

    integer                :: nSources

    type(OmProgressT)      :: reporter    ! ... of progress to GUI
    character(*) , parameter :: routine="omsource_old.omsource_save"

    if (.not.associated(sources)) return ! don't bother
    nSources = size(sources)

    sources%apertureRadius = sources%apertureRadius *  real(myImageWindow%bx) 

    ! notify the GUI that we are busy...
    call construct(reporter, label="Saving source list- please wait", steps = 2)
    call setProgress(reporter, 1)



    ! Create the source-list table
    call CreateSourceListFile(fileName,         &
                              imageSet,         &
                              nSources,         &
                              sources,          &
                              iw%plateScale,    &
                              myApertureRadius / real(myImageWindow%bx) , &
                              myBgrMode,        &
                              3.0,              &
                              myExposure,       & 
                              iw%x0,            &
                              iw%y0,            & 
                              iw%width,         &
                              iw%height) 
    call AddMagnitudesToTable(fileName, sources)     
    call setProgress(reporter, 2)

    call destruct(reporter)

end subroutine omsource_save

!****************************************************************
! Info: computes all properties of the sources specified
! Note: * Each input source is 'empty' save the (xCoord,yCoord) properties
!       * 'trimage' argument is a residual, and should be removed
!****************************************************************
subroutine computeParameters(sources)
    
    type(SourceParameters),pointer    :: sources(:)

    integer(int32)                     :: nSources,nx, ny
   
    
   
    nx = size(myimage, 1)
    ny = size(myimage, 2)
    nSources = size(sources)

    
    call ComputeSourcePositions(sources)


    call omsource_photometry(myimage, nSources, sources) ! see below
    

    ! Compute the RA and Declination of each source
    call setCelestial(myCelestial, sources)
end subroutine computeParameters







!****************************************************************
! Info: computes all properties of the sources specified
! Note: * Each input source is 'empty' save the (xCoord,yCoord) properties
!       * 'trimage' argument is a residual, and should be removed
!****************************************************************
subroutine ComputeSourcePositions(sources)
    
    type(SourceParameters),pointer    :: sources(:)

    integer(int32)                     :: i,nSources,nx, ny
    integer(int32)                     :: i0, j0
    integer(kind=int16) :: sourceNumber = 1
   
    nx = size(myImage, 1)
    ny = size(myImage, 2)
    nSources = size(sources)

   

    ! Compute the centre coordinates of the new sources
    do i = 1, nSources
       
        call computeCentreOfSource(nx,ny, &
              sources(i)%xPos, sources(i)%yPos, &
              i0, j0)
        sources(i)%iMin = max(i0 - 3, 0)
        sources(i)%iMax = min(i0 + 3, nx - 1)
        sources(i)%jMin = max(j0 - 3, 0)
        sources(i)%jMax = min(j0 + 3, ny - 1)

        call ComputePointSourceParameters(sourceNumber, nx - 1, ny - 1, myImage, &
                i0, j0, sources(i))

   
    end do

    
end subroutine ComputeSourcePositions



!************************************************************
! Subroutine photometry
!***********************************************************
subroutine omsource_photometry(image, nSources,  sources)
    real(single),           pointer    :: image(:, :)
    integer(int32),         intent(in) :: nSources
    type(SourceParameters),pointer    :: sources(:)
  

    integer(int32)         :: i
    integer(int32)         :: nx, ny
    integer(int32)         :: bin(2)
    type(SourceParameters), pointer :: s 
    integer(int16), pointer :: levelImage(:, :) => Null()

    real(single) :: corrBackgroundCountRateErr
    nx = size(image, 1)
    ny = size(image, 2)

    ! initialise defaults:
    do i = 1, nSources
        s=>sources(i)
     
        s%innerBckRadius = myBgrRadiusIn
        s%outerBckRadius = myBgrRadiusOut
        s%exposure       = myObsDetails%exposure
        
        s%extensionFlag  = 0
    end do

    ! Do aperture photometry on each sources
    bin = (/myImageWindow%bx, myImageWindow%by/)
!  write(*, *) "omsourceold line 1479 inner=",   sources(1)%innerBckRadius 
! write(*, *) "omsourceold line 1479 outer=",sources(1)%outerBckRadius 


    levelImage => Allocate2DInt16(0, nx - 1, 0, ny - 1, "LevelImage")
!write(*, *) "Omsource-old Line 1488 my aperture radius=", myApertureRadius / real(binx)
    call AperturePhotometry(nx, ny, myImage, levelImage, &
          myQuality, myapertureRadius / real(binX),  nSources, sources)                  
    sources%apertureRadius = sources%apertureRadius / real(binX)
!  write(*, *) "omsourceold line 1487 inner=",   sources(1)%innerBckRadius 
! write(*, *) "omsourceold line 1487 outer=",sources(1)%outerBckRadius 

  

    deallocate(levelImage)



    ! Convert counts to count rates, and set the magnitude
    do i = 1, nSources
        s=>sources(i)

        call ComputeMagnitude(s%corrCountRate, s%corrCountRateErr, &
                       s%corrBackgroundCountRate, corrBackgroundCountRateErr, &
                       s%xPos, s%yPos,             &
                       myScaleCounts, s%extensionFlag, &
                       myCalRadius, myApertureRadius, &
                       myFrameTime, myDeadFraction, &
                       s%magnitude, s%magnitudeErr, s%psf2Correction)

    end do

!  write(*, *) "omsourceold line 1510 inner=",   sources(1)%innerBckRadius 
! write(*, *) "omsourceold line 1510 outer=",sources(1)%outerBckRadius 



end subroutine omsource_photometry
!************************************************************

   ! Info: find a more accurate centre pixel for (xPos,yPos)
!***********************************************************
subroutine computeCentreOfSource(nx, ny, xPos, yPos, i0, j0)
    integer(int32), intent(in) :: nx, ny

     
    real(single), intent(in) :: xPos, yPos

    integer(int32), intent(out) :: i0, j0
    integer(int32) :: il, iu, jl, ju

    integer(int32) :: i, j
    real(single) :: maximum

      ! current pixel centre

    i0 = int(xPos - 0.5)
    j0 = int(yPos - 0.5)

    ! now find a more accurate centre for the image
    il = i0 - 5
    il = max (il, 0)
    iu = i0 + 5
    iu = min (iu, nx - 1)

    jl = j0 - 5
    jl = max (jl, 0)
    ju = j0 + 5
    ju = min (ju, ny - 1)
    maximum = 1e-6

    do i = il, iu
        do j = jl, ju
            if (myImage(i, j) .gt. maximum) then
                maximum = myImage(i, j)
                i0 = i
                j0 = j
            end if
        end do
    end do

end subroutine computeCentreOfSource
!***************************************
!
!***************************************
subroutine construct_od(o, set)
    type(ObsDetails),  intent(out)    :: o
    type(DataSetT),    intent(in)     :: set
      
    call GetStringAttribute(set, "TELESCOP", o%telescope)
    call GetStringAttribute(set, "OBSERVER", o%observer)
    call GetStringAttribute(set, "OBJECT",   o%object)
    call GetStringAttribute(set, "FILTER",   o%filter)
    call GetStringAttribute(set, "INSTRUME", o%instrument)
    call GetStringAttribute(set, "OBS_ID",   o%obsID)
    call GetStringAttribute(set, "EXP_ID",   o%expID)
    call GetStringAttribute(set, "DATE-OBS", o%startObs)
    call GetStringAttribute(set, "DATE-END", o%endObs)
    o%exposure = GetReal32Attribute(set, "EXPOSURE")
end subroutine construct_od
!***************************************
!
!***************************************
subroutine destruct_od(o)
    type(ObsDetails), intent(inout)    :: o
    o%telescope=''
    o%exposure=0.
end subroutine destruct_od
!***************************************
!
!***************************************
subroutine construct_iw(w, set)
    type(ImageWindow), intent(out)    :: w
    type(DataSetT),    intent(in)     :: set
      
    type(TableT)                      :: t
    type(ColumnT)                     :: c
    integer(int32), pointer            :: ptr(:) ! release on table release
    type(DataSetT) :: wdxDataSet      

    t = table(set, "MODES")
    c = column(t, "WINDOWX0", read)
    ptr => int32Data(c)
    w%x0  = ptr(0)
    c = column(t, "WINDOWY0", read)
    ptr => int32Data(c)
    w%y0  = ptr(0)
    c = column(t, "WINDOWDX", read)
    ptr => int32Data(c)
    w%width  = ptr(0)
    c = column(t, "WINDOWDY", read)
    ptr => int32Data(c)
    w%height  = ptr(0)
    myExposure = GetReal32Attribute(set, "EXPOSURE", 0.0)  
   

    call release(t)

    ! compute binning
    w%binBpe = booleanAttribute(set, "BINBPE")
    w%bx     = 2**int32Attribute(set, "BINAX1")
    w%by     = 2**int32Attribute(set, "BINAX2")
    if (w%binBpe) then
        w%bx=w%bx*2
        w%by=w%by*2
    end if
    binX = w%bx
    binY = w%by
    w%filter  = stringAttribute(set, "FILTER")
    w%datamode= stringAttribute(set,"DATAMODE")


    ! CCD frame-time and dead-fraction
    
    if (.not. hasAttribute(set,  "FRAMTIME")) then
        wdxDataSet = dataSet(stringParameter("wdxset"), read)


        call getFrameTime(wdxDataSet, w%frameTime, w%deadFraction, w%binBpe)
        call release(wdxDataSet)
        call setAttribute(set, "FRAMTIME", w%FrameTime, "millisecs", "CCD frame time")
        call setAttribute(set, "DEADFRAC", w%DeadFraction, "", "CCD dead fraction")
        w%frameTime = w%frameTime / 1000.0
       
    else     
        w%frameTime = real64Attribute(set, "FRAMTIME") / 1000.0
        w%deadFraction = real64Attribute(set, "DEADFRAC")

    end if 
end subroutine construct_iw
!***************************************
!
!***************************************
subroutine destruct_iw(w)
    type(ImageWindow), intent(inout)    :: w
    w%filter=''
    w%datamode=''
end subroutine destruct_iw



subroutine omsource_computeparameters(sources)
    type(SourceParameters),  pointer    :: sources(:)

    call ComputeParameters(sources)
end subroutine omsource_computeparameters





!*************************************************
!
! Returns the background mode
!*************************************************

real(single) Function omsource_GetBackgroundMode()
 omsource_GetBackgroundMode = myBgrMode
    
end function omsource_GetBackgroundMode

end module omsource_old










